if (ZOMGBuffTehRaid) then
	z:Print("Installation error, duplicate copy of ZOMGBuffs_BuffTehRaid (Addons\ZOMGBuffs\ZOMGBuffs_BuffTehRaid and Addons\ZOMGBuffs_BuffTehRaid)")
	return
end

local L = AceLibrary("AceLocale-2.2"):new("ZOMGBuffTehRaid")
local R = AceLibrary("ZOMGReagents-1.0")
local BS = AceLibrary("Babble-Spell-2.2")
local roster = AceLibrary("Roster-2.1")
local playerClass
local template
--local mainSingle, mainGroup

local wow24 = GetBuildInfo() < "1.0.0" or GetBuildInfo() >= "2.4.0"

local z = ZOMGBuffs
local zg = z:NewModule("ZOMGBuffTehRaid")
ZOMGBuffTehRaid = zg

local SM = z.SharedMedia

z:CheckVersion("$Revision: 65303 $")

local new, del, deepDel, copy = z.new, z.del, z.deepDel, z.copy

local function getOption(k)
	return zg.db.char[k]
end

local function setOption(k, v, update)
	if (not v or v == 0) then
		zg.db.char[k] = nil
	else
		zg.db.char[k] = v
	end
	if (update) then
		z:CheckForChange(zg)
	end
end

local function getPrelude(k)
	return zg.db.char.rebuff[k] or 0
end

local function setPrelude(k, v)
	if (not v or v == 0) then
		zg.db.char.rebuff[k] = nil
	else
		zg.db.char.rebuff[k] = v
	end
	z:CheckForChange(zg)
end

do
	local function trackerDisabled()
		return not zg.db.char.tracker
	end
	
	zg.consoleCmd = "ZOMGBuffTehRaid"
	zg.options = {
		type = 'group',
		order = 2,
		name = "|cFFFF8080Z|cFFFFFF80O|cFF80FF80M|cFF8080FFG|cFFFFFFFFBuffTehRaid|r",
		desc = L["Group Buff Configuration"],
		handler = zg,
		args = {
			groups = {
				type = 'group',
				name = L["Groups"],
				desc = L["Select the groups to buff"],
				order = 5,
				hidden = function() return not zg:IsModuleActive() end,
				args = {
					auto = {
						type = 'toggle',
						name = L["Auto-Assign"],
						desc = L["Auto assign sensible group assignment based the order of your name alphabilically compared to others of your class. All going well, and all using ZOMGBuffs and everyone should end up with different assignments without need for discussion"],
						get = getOption,
						set = function(k,v) setOption(k,v) zg:AutoGroupAssignment() end,
						passValue = "autogroups",
						order = 1,
					},
					space = {
						type = 'header',
						name = " ",
						order = 2,
					},
				},
			},
			template = {
				type = 'group',
				name = L["Templates"],
				desc = L["Template configuration"],
				order = 50,
				hidden = function() return not zg:IsModuleActive() end,
				args = {
				}
			},
			spacer = {
				type = 'header',
				name = " ",
				hidden = function() return not zg:IsModuleActive() end,
				order = 51,
			},
			tracker = {
				type = 'group',
				name = L["Tracker"],
				desc = L["Tracker Icon for single target exclusive buffs"],
				hidden = function() return not zg:IsModuleActive() end,
				order = 100,
				args = {
					tracker = {
						type = 'toggle',
						name = L["Enable"],
						desc = L["Create a tracking icon for certain exclusive spells (Earth Shield, Fear Ward). Note that the icon can always display the correct status of the spell, but if you change targets in combat then the click action will be to the player who it was last set to before entering combat"],
						get = getOption,
						set = setOption,
						passValue = "tracker",
						order = 1,
					},
					sound = {
						type = 'text',
						name = L["Sound"],
						desc = L["Select a soundfile to play when player's tracked buff expires"],
						get = getOption,
						set = function(k,v)
							setOption(k,v)
							PlaySoundFile(SM:Fetch("sound", v))
						end,
						validate = SM:List("sound"),
						disabled = function() return not SM end,
						passValue = "tracksound",
						order = 20,
					},
					spacer = {
						type = 'header',
						name = " ",
						order = 100,
					},
					scale = {
						type = 'range',
						name = L["Scale"],
						desc = L["Adjust the scale of the tracking icon"],
						func = timeFunc,
						get = getOption,
						set = function(k,v)
							setOption(k,v)
							if (zg.trackIcon) then
								zg.trackIcon:SetPosition()
							end
						end,
						disabled = trackerDisabled,
						passValue = "trackerscale",
						min = 0,
						max = 2,
						isPercent = true,
						step = 0.05,
						order = 101,
					},
					reset = {
						type = 'execute',
						name = L["Reset"],
						desc = L["Reset the position of the tracker icon"],
						order = 102,
						disabled = trackerDisabled,
						func = function() zg.db.char.postracker = nil end,
					}
				}
			},
			behaviour = {
				type = 'group',
				name = L["Behaviour"],
				desc = L["Group buffing behaviour"],
				order = 201,
				hidden = function() return not zg:IsModuleActive() end,
				args = {
					groupcast = {
						type = 'range',
						name = L["Minimum Group"],
						desc = L["How many players of a group must need a buff before the group version is cast"],
						get = getOption,
						set = function(k,v) setOption(k,v,true) end,
						passValue = "groupcast",
						order = 99,
						min = 1,
						max = 5,
						step = 1,
					},
					rebuff = {
						type = 'range',
						name = L["Expiry Prelude"],
						desc = L["Default rebuff prelude for all group buffs"],
						func = timeFunc,
						order = 2,
						get = getPrelude,
						set = setPrelude,
						passValue = "default",
						min = 0,
						max = 15 * 60,
						step = 5,
						bigStep = 60,
						order = 100,
					},
				},
			},
		}
	}
end

do
	local disfunc = function() return zg.db.char.autogroups end
	local getG = function(i) return zg.db.char.groups[i] end
	local setG = function(i,n) zg.db.char.groups[i] = n and true or nil z:CheckForChange(zg) end
	local args = zg.options.args.groups.args
	for i = 1,8 do
		args[i] = {
			type = 'toggle',
			name = format(L["Group %d"], i),
			desc = L["Enable buffing of this group"],
			order = i + 10,
			get = getG,
			set = setG,
			disabled = disfunc,
			passValue = i,
		}
	end
end

local function getOption(v)
	return zg.db.char[v]
end

local function setOption(v, n)
	zg.db.char[v] = n
end

-- ColourSpellFromKey(key)
local function ColourSpellFromKey(key)
	local name = key.name or key.list[1]
	local colour = z:HexColour(unpack(key.colour or {1, 1, 1}))

	if (wow24) then
		local icon = BS:GetSpellIcon(key.list[1])
		if (icon) then
			return format("|T%s:14|t%s%s|r", colour, name), colour
		end
	end

	return format("%s%s|r", colour, name), colour
end

-- ToggleKeyType
local function ToggleKeyType(keytype, onoff)
	if (onoff == nil) then
		onoff = not template[keytype]
	end

	if (onoff) then
		local dup = zg.buffs[keytype].dup
		if (dup) then
			for j,z in pairs(zg.buffs) do
				if (z.dup == dup) then
					zg:ModifyTemplate(j, nil)
				end
			end
		end

		zg:ModifyTemplate(keytype, true)
	else
		zg:ModifyTemplate(keytype, nil)
	end
	z:SetupForSpell()
	zg:CheckBuffs()
end

-- OnModifyTemplate
function zg:OnModifyTemplate(key, value)
	local buff = self.buffs[key]
	if (buff) then
		if (buff.limited) then
			if (value) then
				if (buff.onEnable) then
					buff.onEnable()
				end

				self:RegisterTickColumn(key)

				if (not template.limited or not template.limited[key]) then
					if (not template.limited) then
						template.limited = new()
					end
					if (not template.limited[key]) then
						template.limited[key] = new()
					end

					if (buff.exclusive) then
						template.limited[key][UnitName("player")] = true
					else
						for unit in roster:IterateRoster() do
							template.limited[key][unit.name] = true
						end
					end
				end

				if (buff.exclusive) then
					if (template.limited and template.limited[key]) then
						local name = next(template.limited[key])
						if (name) then
							self:AddSpellTracker(key, name)
						end
					end
				end
			else
				self:UnregisterTickColumn(key)
				self:StopSpellTracker(key)
			end
		end
	end
end

-- SetMenu
function zg:MakeSpellOptions()
	if (self.options.args.singlespells or self.options.args.groupspells) then
		return
	end

	local groupArgs, singleArgs

	local list = self:SortedBuffList()
	for i,key in ipairs(list) do
		local info = self.buffs[key]
		if (GetSpellCooldown(info.list[1])) then
			local name = info.name or info.list[1]
			local cName = ColourSpellFromKey(info)
			local menu = {
				type = "group",
				name = cName,
				desc = cName,
				order = i,
				isChecked = function(k) return template[key] end,
				onClick = ToggleKeyType,
				passValue = key,
				args = {
					header = {
						type = "header",
						name = cName,
						order = 1,
					},
					rebuff = {
						type = "range",
						name = L["Expiry Prelude"],
						desc = format(L["Rebuff prelude for %s (0=Module default)"], cName),
						func = timeFunc,
						order = 2,
						get = getPrelude,
						set = setPrelude,
						passValue = key,
						min = 0,
						max = 30 * 60,
						step = 15,
						bigStep = 60,
						order = 2,
					}
				}
			}

			if (info.limited) then
				if (not singleArgs) then
					local menu = {
						type = 'group',
						name = L["Single Spells"],
						desc = L["Single spell configuration"],
						order = 2,
						args = {
							spacer = {
								type = 'header',
								name = " ",
								order = 999,
							},
							reset = {
								type = 'toggle',
								name = L["Reset on Clear"],
								desc = L["If noone is selected for this buff when you disable it, then the next time it is enabled, everyone will default to ON. If disabled, the last settings will be remembered"],
								get = getOption,
								set = setOption,
								passValue = "resetOnClear",
								order = 1000,
							}
						}
					}
					self.options.args.singlespells = menu
					singleArgs = menu.args
				end

				singleArgs[name] = menu
			else
				if (not groupArgs) then
					local menu = {
						type = 'group',
						name = L["Group Spells"],
						desc = L["Group spell configuration"],
						order = 1,
						args = {
						}
					}
					
					self.options.args.groupspells = menu
					groupArgs = menu.args
				end

				groupArgs[name] = menu
			end
		end
	end
end

-- TickInitForTemplate
function zg:TickInitForTemplate()
	self:MakeTickLookup()

	for Type,buff in pairs(self.buffs) do
		if (buff.limited) then
			if (template[Type]) then
				if (buff.onEnable) then
					buff.onEnable()
				end

				self:RegisterTickColumn(Type)

				if (buff.exclusive) then
					if (template.limited and template.limited[Type]) then
						local name = next(template.limited[Type])
						if (name) then
							self:AddSpellTracker(Type, name)
						end
					end
				end
			else
				self:UnregisterTickColumn(Type)
				self:StopSpellTracker(Type)
			end
		end
	end

	if (z.icon and z.members) then
		z:DrawGroupNumbers()
		z:DrawAllCells()
	end
end

-- SelectTemplate
function zg:OnSelectTemplate(templateName)
	template = self.db.char.templates.current
	self:TickInitForTemplate()
end

-- GetBuffedMembers
function zg:GetBuffedMembers()
	local temp = new()
	local totals = new(0, 0, 0, 0, 0, 0, 0, 0)
	local groupTotal = new(0, 0, 0, 0, 0, 0, 0, 0)
	local groupPresent = new(0, 0, 0, 0, 0, 0, 0, 0)
	local skipGroups = new()
	local dbGroups = self.db.char.groups
	local minTimeLeft
	local totalMembers, totalPresent = 0, 0
	local anyBlacklisted

	for unit in roster:IterateRoster() do
		local subgroup = z:GetGroupNumber(unit.unitid)
		local unitid = unit.unitid
		local grp = subgroup
		anyBlacklisted = anyBlacklisted or z:IsBlacklisted(unit.name)
		totalMembers = totalMembers + 1
		groupTotal[grp] = groupTotal[grp] + 1

		if (dbGroups[grp]) then
			local pvpBlock = (z.db.profile.skippvp and UnitIsPVP(unitid)) and not UnitIsPVP("player")
			local present = UnitIsConnected(unitid) and UnitCanAssist("player", unitid) and not pvpBlock
			local absent						-- They're not in zone, afk, or offline
			if (not present and z.db.profile.ignoreabsent and z.db.profile.waitforclass) then
				if (pvpBlock) then
					absent = true
				elseif (not UnitIsConnected(unitid)) then
					absent = true
				elseif (UnitIsAFK(unitid)) then
					absent = true
				else
					local inZone = true
					if (id) then
						local _, _, _, _, _, _, zone = GetRaidRosterInfo((strmatch(unitid, "(%d+)") or 0) + 0)
						if (zone and zone ~= playerZone) then
							absent = true
						end
					end
				end
			end

			if (present) then
				local foundBuffs
				local manaUser = unit.class ~= "ROGUE" and unit.class ~= "WARRIOR"
				groupPresent[grp] = groupPresent[grp] + 1
				totalPresent = totalPresent + 1
				
				if (not manaUser) then
					foundBuffs = new()
				end

				totals[grp] = totals[grp] + 1
				for i = 1,40 do
					local name, rank, buff, count, max, dur = UnitBuff(unitid, i)
					if (not name) then
						break
					end

					local buff = self.lookup[name]
					if (buff) then
						if (template[buff.type]) then
							if (not manaUser) then
								foundBuffs[buff.type] = true
							end

							local requiredTimeLeft = (self.db.char.rebuff and self.db.char.rebuff[buff.type]) or self.db.char.rebuff.default
							local needsBuff = not buff.onlyManaUsers or manaUser
							if (not needsBuff or (not requiredTimeLeft or not dur or dur > requiredTimeLeft)) then
								if (not temp[buff.type]) then
									temp[buff.type] = new(0, 0, 0, 0, 0, 0, 0, 0)
								end
								temp[buff.type][grp] = temp[buff.type][grp] + 1
							end

							if (needsBuff and dur and (not minTimeLeft or dur - requiredTimeLeft < minTimeLeft)) then
								minTimeLeft = dur - requiredTimeLeft
							end
						end
					end
				end

				if (not manaUser) then
					for key,info in pairs(self.buffs) do
						if (info.onlyManaUsers and not foundBuffs[key]) then
							if (not temp[key]) then
								temp[key] = new(0, 0, 0, 0, 0, 0, 0, 0)
							end
							temp[key][grp] = temp[key][grp] + 1
						end
					end
					del(foundBuffs)
				end

			elseif (absent) then
				totalPresent = totalPresent + 1
				groupPresent[grp] = groupPresent[grp] + 1
				totals[grp] = totals[grp] + 1
				for code,info in pairs(self.buffs) do
					if (not temp[code]) then
						temp[code] = new(0, 0, 0, 0, 0, 0, 0, 0)
					end
					temp[code][grp] = temp[code][grp] + 1
				end
			end
		end
	end

	z.waitingForClass = nil
	if (z.db.profile.waitforclass) then
		for i = 1,8 do
			if (groupTotal[i] > 0 and groupPresent[i] < groupTotal[i]) then
				z.waitingForClass = (z.waitingForClass or "") .. ((z.waitingForClass and ", ") or "") .. format(L["Group %d"], i)
				skipGroups[i] = true
			end
		end
	end

	del(groupPresent)
	del(groupTotal)

	return temp, totals, minTimeLeft, totalPresent / totalMembers, skipGroups, anyBlacklisted
end

-- FindUnitInRangeMissing
function zg:FindUnitInRangeMissing(group, typ)
	local t = self.buffs[typ]
	if (not t) then
		error(format("Unrecognised type %q", tostring(typ)), 2)
	end
	local buffEveryone = not t.onlyManaUsers
	local list = t.list
	local rangeCheckSpell = list[1]
	local limitedPeople = t.limited and template.limited and template.limited[typ]

	local requiredTimeLeft = (self.db.char.rebuff and self.db.char.rebuff[typ]) or self.db.char.rebuff.default

	for unit in roster:IterateRoster() do
		if (buffEveryone or (unit.class ~= "ROGUE" and unit.class ~= "WARRIOR")) then
			if (not t.limited or (limitedPeople and limitedPeople[unit.name])) then
				local subgroup = z:GetGroupNumber(unit.unitid)
				if (subgroup == group and not z:IsBlacklisted(unit.name)) then
					local unitid = unit.unitid
					if (UnitIsConnected(unitid) and UnitCanAssist("player", unitid) and IsSpellInRange(rangeCheckSpell, unitid) == 1 and ((not z.db.profile.skippvp or not UnitIsPVP(unitid)) or UnitIsPVP("player"))) then
						local got
						for i = 1,40 do
							local name, rank, buff, count, max, dur = UnitBuff(unitid, i)
							if (not name) then
								break
							end

							local buff = self.lookup[name]
							if (buff and buff.type == typ) then
								if (dur and dur < requiredTimeLeft) then
									return unitid
								end
								got = true
								break
							end
						end

						if (not got) then
							return unitid
						end
					end
				end
			end
		end
	end
end

-- CheckBuffs
function zg:CheckBuffs()
	if (not self.db or not template or not z:CanCheckBuffs()) then
		return
	end

	-- First build a list of total group counts, and total numbers of players who have each buff in each group
	local temp, totals, minTimeLeft, percentPresent, skipGroups, anyBlacklisted = self:GetBuffedMembers()
	local skipClassBuffs = z.db.profile.singlesInBG and select(2, IsInInstance()) == "pvp"

	if (z.db.profile.waitforraid > 0 and not skipClassBuffs) then
		-- See if enough of raid present
		if (percentPresent < z.db.profile.waitforraid) then	-- Wait for % of raid before buffing
			z.waitingForRaid = floor(percentPresent * 100)
			self:ScheduleEvent("ZOMGBuffTehRaid_CheckBuffs", self.CheckBuffs, 5, self)
			return
		end
	end
	z.waitingForRaid = nil

	-- Now go through this and find anything we need to buff
	local dbGroups = self.db.char.groups
	local any
	for i = 1,8 do								-- i == group
		if (dbGroups[i] and not skipGroups[i]) then
			local players = totals[i]
			if (players > 0) then
				for k,v in pairs(template) do			-- k = spellType (INT, STA etc.), v = buff data for that type
					if (k ~= "modified" and k ~= "state" and k ~= "limited") then
						local typeSpec = self.buffs[k]
						local start, dur = GetSpellCooldown(typeSpec.list[1])
						if (start == 0 or GetTime() - 0.5 > start + dur) then
							local gotBuff = (temp[k] and temp[k][i]) or 0
							if (not gotBuff or gotBuff < players) then
								local missingBuff = players - gotBuff

								local unit = self:FindUnitInRangeMissing(i, k)
								if (unit) then
									local buffSpec = typeSpec.list
									local t1, t2
									if (buffSpec[2]) then
										t1, t2 = IsUsableSpell(buffSpec[2])
									end
									if (skipClassBuffs or (missingBuff < self.db.char.groupcast or not (t1 or t2))) then
										-- Do single buff
										z:Notice(format(L["%s needs %s"], z:ColourUnit(unit), typeSpec.name or buffSpec[1]), "buffreminder")
										z:SetupForSpell(unit, buffSpec[1], self)
										any = true
										break
									else
										-- Do group buff
										z:Notice(format(L["%s needs %s"], z:ColourGroup(i), typeSpec.name or buffSpec[1]), "buffreminder")
										z:SetupForSpell(unit, buffSpec[2], self)
										any = true
										break
									end
								--else
									--error("Know we're supposed to buff somoene, but can't find anyone atm")
								end
							end
						else
							-- Set min time left to cooldown end for this spell
							minTimeLeft = min((start + dur) - GetTime(), minTimeLeft or 0)
						end
					end
				end
				if (any) then
					break
				end
			end
		end
	end

	if (not any and z.db.profile.buffpets) then
		-- Catch any pets that missed the group buffing
		for unit in roster:IterateRoster(true) do
			local unitid = unit.unitid
			local ownerunitid
			if (unitid == "pet") then
				ownerunitid = "player"
			else
				ownerunitid = unitid:gsub("pet", "")
			end
			local ownerunit = roster:GetUnitObjectFromUnit(ownerunitid)
			if (ownerunit and unit.class == "PET" and UnitIsVisible(unitid) and UnitCanAssist("player", unitid) and CheckInteractDistance(unitid, 1)) then
				local subgroup = z:GetGroupNumber(ownerunit)

 				if (self.db.char.groups[subgroup]) then
					for i = 1,40 do
						local name, rank, buff, count, max, dur = UnitBuff(unitid, i)
						if (not name) then
							break
						end

						local buff = self.lookup[name]
						if (buff) then
							if (template[buff.type]) then
								local requiredTimeLeft = (self.db.char.rebuff and self.db.char.rebuff[buff.type]) or self.db.char.rebuff.default
								if ((not requiredTimeLeft or not dur or dur > requiredTimeLeft) and (buff.onlyManaUsers and (unit.class == "ROGUE" or unit.class == "WARRIOR"))) then
									z:Notice(format(L["%s needs %s"], z:ColourUnit(unitid), buff.name), "buffreminder")
									z:SetupForSpell(unit, buff.list[1], self)
									any = true
									break
								end

								if (dur and (not minTimeLeft or dur - requiredTimeLeft < minTimeLeft)) then
									minTimeLeft = dur - requiredTimeLeft
								end
							end
						end
					end
					if (any) then
						break
					end
				end
			end
		end
	end

	self:CancelScheduledEvent("ZOMGBuffTehRaid_CheckBuffs")
	if (anyBlacklisted) then
		minTimeLeft = 1.5
	end
	if (any) then
		z.waitingForRaid = nil
		z.waitingForClass = nil
	else
		if (z.waitingForClass) then
			self:ScheduleEvent("ZOMGBuffTehRaid_CheckBuffs", self.CheckBuffs, 5, self)
		else
			self:ScheduleEvent("ZOMGBuffTehRaid_CheckBuffs", self.CheckBuffs, minTimeLeft or 60, self)
		end
	end

	deepDel(temp)
	deepDel(groups)
	del(totals)
	del(skipGroups)
end

-- UNIT_AURA
function zg:UNIT_AURA(unit)
	if (UnitInParty(unit) or UnitInRaid(unit)) then
		-- If we have a buff loaded up, then we need to see if another of our class has buffed this person
		z:CheckForChange(self)
	end
end

-- GetActions
function zg:GetActions()
	if (self.buffs and not self.actions) then
		self.actions = {}

		local list = self:SortedBuffList()
		for i,key in ipairs(list) do
			local buff = self.buffs[key]

			if (GetSpellCooldown(buff.list[1])) then
				if (buff.keycode) then
					if (buff.list[2]) then
						for j = 1,#buff.list do
							tinsert(self.actions, {name = buff.list[j], type = buff.keycode..j})
						end
					else
						tinsert(self.actions, {name = buff.list[1], type = buff.keycode})
					end
				end
			end
		end

		del(list)
	end
	return self.actions
end

-- ResetActions
function zg:ResetActions()
	self.actions = nil
end

-- RebuffQuery
function zg:RebuffQuery(unit)
	if (self.db and template and UnitIsConnected(unit) and UnitCanAssist("player", unit) and not UnitIsDeadOrGhost(unit)) then
		local got = new()

		for i = 1,40 do
			local name, rank, buff, count, max, dur = UnitBuff(unit, i)
			if (not name) then
				break
			end

			local info = self.lookup[name]
			if (info) then
				got[info.type] = true
			end
		end

		local c = select(2, UnitClass(unit))
		local manaUser = c ~= "ROGUE" and c ~= "WARRIOR"
		local limited = template.limited
		for key,info in pairs(self.buffs) do
			if (template[key] and not got[key]) then
				if (info.onlyManaUsers or manaUser) then
					if (info.limited) then
						if (limited and limited[key] and limited[key][UnitName(unit)]) then
							del(got)
							return true
						end
					else
						del(got)
						return true
					end
				end
			end
		end

		del(got)
	end

	return
end

-- OnModuleInitialize
function zg:OnModuleInitialize()
	playerClass = select(2, UnitClass("player"))

	if (playerClass == "PRIEST") then
		local t1, t2 = IsUsableSpell(BS["Divine Spirit"])
		self.buffs = {
			STA = {
				o = 1,
				name = SPELL_STAT3_NAME,
				list = {BS["Power Word: Fortitude"], BS["Prayer of Fortitude"]},
				reagents = {nil, {
						[1] = R["Holy Candle"],
						[2] = R["Sacred Candle"],
						[3] = R["Sacred Candle"],
					},
				},
				required = true,
				colour = {0.7, 0.7, 1},
				keycode = "stamina",
			},
			SPIRIT = {
				o = 2,
				name = SPELL_STAT5_NAME,
				list = {BS["Divine Spirit"], BS["Prayer of Spirit"]},
				reagents = {nil, R["Sacred Candle"]},
				required = t1 or t2,
				onlyManaUsers = true,
				colour = {0.3, 0.5, 1},
				keycode = "spirit",
			},
			SHADOWPROT = {
				o = 3,
				list = {BS["Shadow Protection"], BS["Prayer of Shadow Protection"]},
				reagents = {nil, R["Sacred Candle"]},
				colour = {0.7, 0.2, 0.7},
				keycode = "shadowprot",
			},
			FEARWARD = {
				o = 4,
				list = {BS["Fear Ward"]},
				colour = {1, 0.7, 0.3},
				limited = true,						-- Allow limited targets config
				exclusive = true,					-- Can only be cast on 1 target
				keycode = "fearward",
			}
		}
		self.reagents = {
			[R["Holy Candle"]] = {0, 1, 500, maxLevel = 59},
			[R["Sacred Candle"]] = {0, 1, 500},
		}

	elseif (playerClass == "DRUID") then
		self.buffs = {
			MARK = {
				o = 1,
				name = L["Mark"],
				list = {BS["Mark of the Wild"], BS["Gift of the Wild"]},
				reagents = {nil, {
						[1] = R["Wild Berries"],
						[2] = R["Wild Thornroot"],
						[3] = R["Wild Quillvine"],
					},
				},
				required = true,
				colour = {0.9, 0.5, 0.9},
				keycode = "mark",
			},
			THORNS = {
				o = 2,
				list = {BS["Thorns"]},
				colour = {0.7, 0.7, 0.3},
				limited = true,						-- Allow limited targets config
				keycode = "thorns",
			}
		}
		self.reagents = {
			[R["Wild Berries"]] = {0, 1, 500},
			[R["Wild Thornroot"]] = {0, 1, 500},
			[R["Wild Quillvine"]] = {0, 1, 500}
		}

	elseif (playerClass == "MAGE") then
		self.buffs = {
			INT = {
				o = 1,
				name = SPELL_STAT4_NAME,
				list = {BS["Arcane Intellect"], BS["Arcane Brilliance"]},
				reagents = {nil, R["Arcane Powder"]},
				required = true,
				onlyManaUsers = true,
				colour = {0.3, 0.3, 1},
				keycode = "int"
			},
			DAMPEN = {
				o = 2,
				list = {BS["Dampen Magic"]},
				dup = 1,
				colour = {0.3, 1, 0.8},
				limited = true,						-- Allow limited targets config
				keycode = "dampen",
			},
			AMPLIFY = {
				o = 3,
				list = {BS["Amplify Magic"]},
				dup = 1,
				colour = {1, 1, 0.25},
				limited = true,						-- Allow limited targets config
				keycode = "amplify",
			}
		}
		self.reagents = {
			[R["Arcane Powder"]] = {(UnitLevel("player") >= 56 and 20) or 0, 1, 500},
		}
	elseif (playerClass == "SHAMAN") then
		self.buffs = {
			EARTHSHIELD = {
				o = 1,
				list = {BS["Earth Shield"]},
				onEnable = function()
					local zs = ZOMGSelfBuffs
					if (zs) then
						zs:ModifyTemplate(BS["Lightning Shield"], nil)
						zs:ModifyTemplate(BS["Water Shield"], nil)
					end
				end,
				colour = {0.7, 0.7, 0.2},
				limited = true,						-- Allow limited targets config
				exclusive = true,					-- Can only be cast on 1 target
				keycode = "earthshield",
			},
			WATERWALK = {
				o = 2,
				list = {BS["Water Walking"]},
				colour = {0.5, 0.5, 1},
				limited = true,						-- Allow limited targets config
				keycode = "water",
			},
			WATERBREATH = {
				o = 2,
				list = {BS["Water Breathing"]},
				colour = {0.2, 0.8, 1},
				limited = true,						-- Allow limited targets config
				keycode = "breath",
			}
		}
	
	elseif (playerClass == "WARLOCK") then
		self.buffs = {
			SEEINVIS = {
				o = 1,
				list = {BS["Detect Greater Invisibility"]},
				alternatives = {BS["Detect Invisibility"], BS["Detect Lesser Invisibility"]},
				colour = {0.7, 0.2, 0.2},
				limited = true,						-- Allow limited targets config
				keycode = "seeinvis",
			},
			WATERBREATH = {
				o = 2,
				list = {BS["Unending Breath"]},
				colour = {0.5, 1, 0.5},
				limited = true,						-- Allow limited targets config
				keycode = "breath",
			}
		}	
	end

	if (not self.buffs) then
		return
	end

	z.SetClickSpells = function(self, cell)
		if (zg.actions) then
			for i,action in ipairs(zg.actions) do
				local Mod, button = z:GetActionClick(action.type)
				if (button) then
					z:SetACellSpell(cell, Mod, button, action.name)
				end
			end
		end
	end

	local defaultForClass = {Default = {}}
	for k,v in pairs(self.buffs) do
		if (v.required) then
			defaultForClass.Default[k] = true
		end
	end

	self.db = z:AcquireDBNamespace("BuffTehRaid")
	z:RegisterDefaults("BuffTehRaid", "char", {
		reagents = {},
		templates = defaultForClass,
		defaultTemplate = "Default",
		groupcast = 2,
		groups = {true, true, true, true, true, true, true, true},
		autogroups = true,
		rebuff = {
			default = 30,
		},
		resetOnClear = true,
		tracker = true,
		trackerscale = 1,
		tracksound = "Gong",
	} )
	z:RegisterChatCommand({"/zomgraid", "/zomgbufftehraid"}, zg.options)
	self.OnMenuRequest = self.options
	z.options.args.ZOMGBuffTehRaid = self.options

	self.lookup = {}
	self.required = 0
	for k,v in pairs(self.buffs) do
		v.type = k
		for i,buff in pairs(v.list) do
			self.lookup[buff] = v
		end
	end

	z.OnCommReceive.AUTOGROUPASSIGNED = function(self, prefix, sender, channel, class, displayList)
		if (not zg.db.char.autogroups and class == playerClass) then
			if (not zg.autoGroupWarned) then
				zg.autoGroupWarned = true
				self:Print(L["Warning: %s has auto-assigned themselves to buff groups %s, but you have the Auto Group Assignment option disabled"], z:ColourUnitByName(sender), table.concat(displayList, ", "))
			end
		end
	end

	for k,buff in pairs(self.buffs) do
		if (buff.limited) then
			self:RegisterTickHandler(k, self, zg.TickColumnCallback, buff.list[1])
		end
	end

	self.OnModuleInitialize = nil
end

-- TickColumnCallback
function zg:TickColumnCallback(unit, key, enable)
	local handler = self.tickHandlers[key]
	if (not handler) then
		error(format("No handler for key %q", tostring(key)), 2)
	end
	local column = self.tickColumnLookup[key]
	if (not column) then
		error(format("No column for key %q", tostring(key)), 2)
	end
	if (column) then
		local name = UnitName(unit)
		if (name) then
			if (enable == nil) then
				local list = template.limited and template.limited[key]
				return (list and list[name] and 1) or nil
			else
				if (enable) then
					self:AddLimitedSpell(name, key)
				else
					self:RemoveLimitedSpell(name, key)
				end
			end
		end
	end
end

-- RegisterTickHandler
function zg:RegisterTickHandler(id, Module, Func, spellname)
	if (not self.tickHandlers) then
		self.tickHandlers = {}
	end
	self.tickHandlers[id] = {module = Module, func = Func, spell = spellname}
end

-- RegisterTickColumn
function zg:RegisterTickColumn(key)
	if (not self:IsTickColumnRegistered(key)) then
		if (not self.tickHandlers[key]) then
			error(format("No handler for key %q", tostring(key)), 2)
		end
		if (not self.tickColumns) then
			self.tickColumns = {}
		end
		tinsert(self.tickColumns, key)
		self.anyTickColumns = true

		self:MakeTickLookup()

		if (z.icon and z.members) then
			z:DrawGroupNumbers()
			z:DrawAllCells()
		end
	end
end

-- MakeTickLookup
function zg:MakeTickLookup()
	if (self.tickColumns) then
		del(self.tickColumnLookup)
		self.tickColumnLookup = new()
		for i,key in ipairs(self.tickColumns) do
			self.tickColumnLookup[key] = i
		end
	end
end

-- RegisterTickColumn
function zg:UnregisterTickColumn(key)
	local index = self:IsTickColumnRegistered(key)
	if (index) then
		tremove(self.tickColumns, index)

		if (z.icon and z.members) then
			z:DrawGroupNumbers()
			z:DrawAllCells()
		end
	end
end

-- IsTickColumnRegistered
function zg:IsTickColumnRegistered(id)
	if (self.tickColumns) then
		for i,key in ipairs(self.tickColumns) do
			if (key == id) then
				return i
			end
		end
	end
end

-- tickOnClick
local function tickOnClick(self, button)
	if (not zg.tickColumns) then
		-- Shouldn't be happening
		return
	end
	local index = self:GetID()
	local parent = self:GetParent()
	local unitid = parent:GetAttribute("unit")
	local u = roster:GetUnitObjectFromUnit(unitid)
	local key = zg.tickColumns[index]
	local handler = zg.tickHandlers[key]
	local buff = zg.buffs[key]
	if (not key) then
		error(format("No key for column %s", tostring(index)))
	end
	if (not handler) then
		error(format("No handler for %q", tostring(key)))
	end
	if (not buff) then
		error(format("No buff for %q", tostring(key)))
	end

	local enable = self:GetChecked() and true or false
	if (not buff.exclusive and button == "LeftButton" and IsAltKeyDown()) then
		local class = u.class
		zg:TickSome(index, function(unit) return unit.class == class end, enable)	-- Tick all of this class

	elseif (not buff.exclusive and u and button == "LeftButton" and IsShiftKeyDown()) then
		local group = u.subgroup
		zg:TickSome(index, function(unit) return unit.subgroup == group end, enable)-- Tick party

	elseif (not buff.exclusive and u and button == "RightButton") then
		zg:TickSome(index, nil, enable)												-- Tick everyone

	elseif (u) then
		zg:TickOne(index, UnitName(u.name), enable)									-- Tick this one
	end

	z:DrawAllCells()
end

-- tickOnEnter
local function tickOnEnter(self)
	local index = self:GetID()
	local parent = self:GetParent()
	local unitid = parent:GetAttribute("unit")
	local u = roster:GetUnitObjectFromUnit(unitid)
	local key = zg.tickColumns[index]
	local handler = zg.tickHandlers[key]
	local buff = zg.buffs[key]

	local done
	for i = 1,20 do
		local help = L:HasTranslation("TICKCLICKHELP"..i) and L["TICKCLICKHELP"..i]
		if (not help) then
			break
		end
		if (not done) then
			done = true
			GameTooltip:SetOwner(self, "ANCHOR_TOP")
			GameTooltip:SetText(format(L["%s on %s"], ColourSpellFromKey(buff), z:ColourUnitByName(UnitName(unitid))))
		end
		help = help:gsub("$class", z:ColourClass(u.class))
		help = help:gsub("$party", z:ColourGroup(u.subgroup))
		GameTooltip:AddLine(help)
	end

	if (done) then
		GameTooltip:Show()
	end
end

-- gttOnLeave
local function gttOnLeave()
	GameTooltip:Hide()
end

-- TickOne
function zg:TickOne(index, unitname, enable)
	local key = self.tickColumns[index]
	if (not key) then
		error(format("No key for column %s", tostring(index)), 2)
	end
	local handler = self.tickHandlers[key]
	if (not handler) then
		error(format("No handler for %s", tostring(key)), 2)
	end

	handler.func(handler.module, unitname, key, enable)
end

-- TickSome
function zg:TickSome(index, matchfunc, enable)
	for unit in roster:IterateRoster() do
		if (not matchfunc or matchfunc(unit)) then
			self:TickOne(index, unit.name, enable)
		end
	end
end

-- CreateTick
function zg:CreateTick()
	local tick = CreateFrame("CheckButton")
	tick:SetHeight(z.db.char.height)
	tick:SetWidth(z.db.char.height)
	tick:SetNormalTexture("Interface\\Buttons\\UI-CheckBox-Up")
	tick:SetPushedTexture("Interface\\Buttons\\UI-CheckBox-Down")
	tick:SetHighlightTexture("Interface\\Buttons\\UI-CheckBox-Highlight", "ADD")
	tick:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check")
	tick:SetDisabledCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check-Disabled")
	tick:RegisterForClicks("LeftButtonUp", "RightButtonUp")

	tick:SetScript("OnClick", tickOnClick)
	tick:SetScript("OnEnter", tickOnEnter)
	tick:SetScript("OnLeave", gttOnLeave)

	return tick
end

-- AquireTick
function zg:AquireTick(cell)
	if (not self.freeTicks) then
		self.freeTicks = {}
	end
	local tick = tremove(self.freeTicks, 1)
	if (not tick) then
		tick = self:CreateTick()
	end

	if (not cell.ticks) then
		cell.ticks = {}
	end
	tinsert(cell.ticks, tick)
	tick:SetParent(cell)
	tick:Show()
	tick:SetID(#cell.ticks)

	if (#cell.ticks == 1) then
		tick:SetPoint("LEFT")
	else
		tick:SetPoint("LEFT", cell.ticks[#cell.ticks - 1], "RIGHT")
	end

	return tick
end

-- KillTick
function zg:KillTick(cell)
	local tick = tremove(cell.ticks, #cell.ticks)
	if (tick) then
		tinsert(self.freeTicks, tick)
		tick:Hide()
		tick:SetParent(nil)
	end
end

-- CheckTickColumns
function zg:CheckTickColumns(cell)
	local count = (self.tickColumns and #self.tickColumns) or 0
	-- Add any more we need
	while (not cell.ticks or #cell.ticks < count) do
		local tick = self:AquireTick(cell)
	end
	-- Remove any to be recycled
	while (cell.ticks and #cell.ticks > count) do
		self:KillTick(cell)
	end

	local b = cell.buff[1]
	if (not b or not b:IsShown()) then
		b = cell.palaIcon[1]
	end
	if (b and b:IsShown()) then
		b:ClearAllPoints()
		if (count == 0) then
			b:SetPoint("TOPLEFT", 0, 0)
		else
			b:SetPoint("TOPLEFT", cell.ticks[count], "TOPRIGHT")
		end
	end

	if (self.tickColumns) then
		-- Go thru the ticks we have and assign them their values
		local unit = cell:GetAttribute("unit")
		local any
		for index,key in ipairs(self.tickColumns) do
			local handler = self.tickHandlers[key]
			if (type(handler) ~= "table" or not handler.func or not handler.module) then
				error(format("No handler for %q", tostring(key)))
			end
			cell.ticks[index]:SetChecked(handler.func(handler.module, unit, key))
			any = true
		end

		if (not any) then
			self.tickColumns = del(self.tickColumns)
		end
	end
end

-- CheckTickTitles
function zg:CheckTickTitles(members)
	local topChild = members:GetAttribute("child1")
	if (topChild) then
		local show = topChild:IsShown()

		if (self.tickColumns) then
			for i,key in ipairs(self.tickColumns) do
				local relative = topChild.ticks and topChild.ticks[i]
				if (relative) then
					local handler = self.tickHandlers[key]
					if (not handler) then
						error(format("No handler for %q", tostring(key)))
					end

					if (not members.limitedIcons) then
						members.limitedIcons = {}
					end
					local icons = members.limitedIcons

					local icon = icons[i]
					if (not icon) then
						icon = members:CreateTexture(nil, "BORDER")
						icons[i] = icon
					end
					icon:SetWidth(z.db.char.height)
					icon:SetHeight(z.db.char.height)
					icon:SetTexture(BS:GetSpellIcon(handler.spell))

					icon:ClearAllPoints()
					icon:SetPoint("BOTTOM", relative, "TOP", 0, (self.db.char.border and 3) or 1)
					if show then
						icon:Show()
					else
						icon:Hide()
					end
				end
			end
			if (members.limitedIcons) then
				for i = #self.tickColumns + 1,#members.limitedIcons do
					members.limitedIcons[i]:Hide()
				end
			end
		else
			if (members.limitedIcons) then
				for i,icon in ipairs(members.limitedIcons) do
					icon:Hide()
				end
			end
		end
	end
end

-- AutoGroupAssignment
function zg:AutoGroupAssignment()
	if (not self.db or not self.db.char.autogroups) then
		return
	end

	local playerName = UnitName("player")
	local count = z.classcount[playerClass]
	local list = new()
	local maxGroup = 1
	if (GetNumRaidMembers() > 0) then
		for i = 1,GetNumRaidMembers() do
			local name, rank, subgroup, level, _, class = GetRaidRosterInfo(i)
			if (class == playerClass) then
				tinsert(list, name)
			end
			if (subgroup > maxGroup) then
				maxGroup = subgroup
			end
		end
	end
	sort(list)

	local myIndex = 1
	for i,name in ipairs(list) do
		if (name == playerName) then
			myIndex = i
			break
		end
	end

	local displayList = new()
	if (maxGroup == 1 or count == 1) then
		for i = 1,8 do
			self.db.char.groups[i] = true
		end
		z:CheckForChange(self)
		return
	end

	local anyChange
	local each = maxGroup / count
	local start = (myIndex - 1) * each + 1
	for groupNo = 1,maxGroup do
		local shouldBuff = groupNo >= start and groupNo < start + each
		if (shouldBuff) then
			tinsert(displayList, tostring(groupNo))
		end
		if (shouldBuff ~= self.db.char.groups[groupNo]) then
			self.db.char.groups[groupNo] = shouldBuff
			anyChange = true
		end
	end
	if (not next(displayList)) then
		for i = 1,8 do
			self.db.char.groups[i] = true
		end
		z:CheckForChange(self)
		return
	end

	if (anyChange) then
		if (GetNumRaidMembers() > 0 and z.db.profile.info) then
			self:Print(L["You are now responsible for Groups %s"], table.concat(displayList, ", ", 1, max(1, #displayList - 1))..((#displayList > 1 and L[" and "]..displayList[#displayList]) or ""))
		end
		z:CheckForChange(self)
		z:SendCommMessage("GROUP", "AUTOGROUPASSIGNED", playerClass, displayList)
	end

	del(list)
	del(displayList)
end

-- CheckRoster
function zg:CheckRoster()
	if (self.db) then
		self:AutoGroupAssignment()
	end
end

-- ShowBuffBar
function zg:ShowBuffBar(cell, name)
	if (self.lookup) then
		local prefix = SecureButton_GetModifierPrefix(cell)
		for button = 1,5 do
			local spell = cell:GetAttribute(prefix, "spell", button)
			if (spell == name) then
				return true
			else
				local find = self.lookup[spell]
				if (find and find.list[2]) then
					if (find.list[1] == name or find.list[2] == name) then
						return true
					end
				end
			end
		end
	end
end

-- GetSpellColour
function zg:GetSpellColour(spellName)
	local def = self.lookup and self.lookup[spellName]
	if (def) then
		return z:HexColour(unpack(def.colour))
	end
end

-- OneOfYours
function zg:OneOfYours(spell)
	return self.lookup[spell]
end

-- SayWhatWeDid
function zg:SayWhatWeDid(icon, spell, name, rank)
	local s = spell or icon:GetAttribute("spell")
	if (s) then
		local found = self.lookup[s]
		if (found) then
			local _, colour = ColourSpellFromKey(found)
			local unit = roster:GetUnitObjectFromName(name)
			if (unit and found.list and spell == found.list[2]) then
				local r = found.reagents and found.reagents[2]
				local subgroup = z:GetGroupNumber(unit.unitid)
				if (r) then
					local reagent
					if (type(r) == "table") then
						if (r[rank]) then
							reagent = r[rank]
						else
							regent = r[#r]
						end
					else
						reagent = r
					end

					local count = GetItemCount(reagent)
					local colourCount
					if (count < 5) then
						colourCount = "|cFFFF4040"
					elseif (count < 10) then
						colourCount = "|cFFFFFF40"
					else
						colourCount = "|cFF40FF40"
					end

					self:Print(L["%s on Group %d (%s%d|r)"], z:LinkSpell(s, colour), subgroup, colourCount, count - 1)
				else
		      		self:Print(L["%s on Group %d"], z:LinkSpell(s, colour), subgroup)
				end
			else
				self:Print(L["%s on %s"], z:LinkSpell(s, colour), z:ColourUnitByName(name))
			end
		end
	end
end

-- OnRegenEnabled
function zg:OnRegenEnabled()
	if (self.stopSpellTracking) then
		self:StopSpellTracker(self.stopSpellTracking)
	end
end

-- CreateTrackIcon
function zg:CreateTrackIcon()
	local iname = "ZOMGTrackIcon"
	local icon = CreateFrame("Button", iname, UIParent, "SecureActionButtonTemplate,ActionButtonTemplate")

	icon:RegisterForDrag(nil)

	icon.icon = getglobal(iname.."Icon")
	icon.border = getglobal(iname.."Border")
	icon.normal = getglobal(iname.."NormalTexture")
	icon.flash = getglobal(iname.."Flash")
	icon.hotkey = getglobal(iname.."HotKey")
	icon.name = getglobal(iname.."Name")
	icon.count = getglobal(iname.."Count")
	icon.cooldown = getglobal(iname.."Cooldown")

	icon.name:ClearAllPoints()
	icon.name:SetPoint("TOPLEFT")
	icon.name:SetPoint("BOTTOMRIGHT", icon, "TOPRIGHT", 0, -20)
	icon.name:SetJustifyV("TOP")
	icon.name:SetNonSpaceWrap(true)
	icon.name:SetFontObject("NumberFontNormal")

	icon.icon:SetTexCoord(0.06, 0.94, 0.06, 0.94)
	icon.border:SetVertexColor(0, 1, 0, 0.6)
	icon.normal:SetVertexColor(1, 1, 1, 0.5)
	icon:SetWidth(36)
	icon:SetHeight(36)
	
	icon.swirl = CreateFrame("Model", nil, icon)
	icon.swirl:SetModel("Interface\\Buttons\\UI-AutoCastButton.mdx")
	icon.swirl:Hide()
	icon.swirl:SetAllPoints()
	icon.swirl:SetSequence(0)
	icon.swirl:SetSequenceTime(0, 0)
	icon.swirl:SetScale(1.4)

	icon:RegisterForClicks("AnyUp")
	icon:RegisterForDrag("LeftButton")
	icon:SetMovable(true)

	-- SmoothColour
	local function SmoothColour(percentage)
		local r, g
		if (percentage < 0.5) then
			g = 2*percentage
			r = 1
		else
			g = 1
			r = 2*(1 - percentage)
		end
		if (r < 0) then r = 0 elseif (r > 1) then r = 1 end
		if (g < 0) then g = 0 elseif (g > 1) then g = 1 end
		return r, g, 0
	end

	function icon:UpdateCooldown()
		local start, duration, enable = GetSpellCooldown(self.spell)
		CooldownFrame_SetTimer(self.cooldown, start, duration, enable)
	end

	function icon:GetSpellFromUnit()
		for i = 1,40 do
			local name, rank, buff, count, maxDuration, duration = UnitBuff(self.target, i)
			if (not name) then
				return
			end
			if (name == self.spell and duration) then
				self.timeLeft = duration
				self.maxTime = maxDuration
				return count, duration, maxDuration
			end
		end
		self.timeLeft = nil
	end
	
	function icon:UpdateAura()
		local count, duration, maxDuration = self:GetSpellFromUnit()
		if (duration) then
			self.timeLeft = duration
			self.stacks = count
			if (not self.initialStacks) then
				self.initialStacks = count
			end
			self.count:SetText(count)

			if (not self.initialStacks or self.initialStacks == 1) then
				self.count:Hide()
			else
				self.count:Show()
				local r, g, b = SmoothColour(self.stacks / self.initialStacks)
				self.count:SetTextColor(r, g, b)
			end
			self.swirl:Hide()
			return
		else
			self.timeLeft = nil
		end

		if (self.stacks and self.stacks > 0) then
			local buff = zg.buffs[self.key]
			z:Notice(format(L["%s has expired on %s"],  ColourSpellFromKey(buff), z:ColourUnitByName(self.target)))
			PlaySoundFile(SM:Fetch("sound", zg.db.char.tracksound))
		end

		self.name:Hide()
		self.stacks = 0
		self.count:SetText("0")
		self.count:SetTextColor(1, 0, 0)
		self.swirl:Show()
	end

	function icon:UpdateTooltip()
		local count, duration, maxDuration = self:GetSpellFromUnit()
		local buff = zg.buffs[self.key]
		local buffColour = z:HexColour(unpack(buff.colour))

		GameTooltip:SetOwner(self, "ANCHOR_TOP")
		GameTooltip:SetText(format("%s %s", z.titleColour, L["Spell Tracker"]), 1, 1, 1)
		GameTooltip:AddLine(format(L["%s on %s"], ColourSpellFromKey(buff), z:ColourUnitByName(self.target)))

		if (duration) then
			local r, g, b
			if (self.stacks and self.initialStacks and self.initialStacks ~= 1) then
				r, g, b = SmoothColour(self.stacks / self.initialStacks)
				GameTooltip:AddLine(format(L["%d of %d stacks remain"], self.stacks or 0, self.initialStacks or 0), r, g, b)
			end
			local r, g, b = SmoothColour(duration / maxDuration)
			GameTooltip:AddLine(format("%s remains", date("%M:%S", duration)), r, g, b)
		else
			if (count) then
				GameTooltip:AddLine(L["Buffed from someone else"], 0.5, 0.5, 0.5)
			else
				GameTooltip:AddLine(L["MISSING!"], 1, 0, 0)
			end
		end

		if (not UnitIsUnit(self.target, self:GetAttribute("unit"))) then
			GameTooltip:AddLine(L["WARNING: The intended target for this icon has changed since you entered combat"], 1, 0.5, 0.5)
		end

		GameTooltip:Show()
	end
	
	icon:SetScript("OnDragStart",
		function(self)
			self:StartMoving()
		end)

	icon:SetScript("OnDragStop",
		function(self)
			self:StopMovingOrSizing()
			zg.db.char.postracker = z:GetPosition(self)
		end)
	
	icon:SetScript("OnUpdate",
		function(self, elapsed)
			if (self.timeLeft) then
				self.timeLeft = self.timeLeft - elapsed
				if (self.timeLeft > 0) then
					self.name:SetText(date("%M:%S", self.timeLeft):gsub("^0", ""))
					local r, g, b = SmoothColour(self.timeLeft / self.maxTime)
					self.name:SetTextColor(r, g, b)
					self.name:Show()
				else
					self.timeLeft = nil
				end
			else
				self.name:Hide()
			end
		end)

	icon:SetScript("OnEvent",
		function(self, event, unit, ...)
			if (event == "UNIT_AURA") then
				if (UnitIsUnit(unit, self.target)) then
					self:UpdateAura()
				end
			elseif (event == "ACTIONBAR_UPDATE_COOLDOWN") then
				self:UpdateCooldown()
			end
		end)
		
	icon:SetScript("OnEnter", icon.UpdateTooltip)
	icon:SetScript("OnLeave", gttOnLeave)

	icon:SetScript("OnShow",
		function(self)
			self:RegisterEvent("UNIT_AURA")
			self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
			self:SetPosition()
			self:UpdateCooldown()
		end)

	icon:SetScript("OnHide",
		function(self)
			self:UnregisterEvent("UNIT_AURA")
			self:UnregisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
		end)

	function icon:SetPosition()
		self:SetPoint("CENTER")
		self:SetScale(zg.db.char.trackerscale)
		z:RestorePosition(self, zg.db.char.postracker)
	end
	
	-- SetTrackerAttributes
	function icon:SetTrackerAttributes(target, key)
		local buff = zg.buffs[key]
		local spell = buff.list[1]
		self.target = target
		self.key = key
		self.spell = spell
		self.initialStacks = nil
		self.count:Hide()
		self.name:Hide()
		self.icon:SetTexture(BS:GetSpellIcon(spell))

		if (not InCombatLockdown()) then
			self:SetAttribute("unit", self.target)
			self:SetAttribute("type", "spell")
			self:SetAttribute("spell", self.spell)
		end
	end

	icon:Hide()

	self.trackIcon = icon
	self.CreateTrackIcon = nil
	return icon
end

-- AddSpellTracker
function zg:AddSpellTracker(key, target)
	if (not self.db.char.tracker) then
		return
	end

	local icon = self.trackIcon
	if (not icon) then
		icon = self:CreateTrackIcon()
	end

	icon:SetTrackerAttributes(target, key)
	icon:UpdateAura()
	icon:Show()
end

-- StopSpellTracker
function zg:StopSpellTracker(key)
	if (self:IsSpellTracking(key)) then
		if (self.trackIcon) then
			if (InCombatLockdown()) then
				self.stopSpellTracking = key
			else
				self.trackIcon:Hide()
			end
		end
	end
end

-- IsSpellTracking
function zg:IsSpellTracking(key)
	local buff = self.buffs[key]
	local icon = self.trackIcon
	return buff and icon and icon.spell and icon.spell == buff.list[1]
end

-- SpellCastSucceeded
function zg:SpellCastSucceeded(spell, rank, target, manual, listClick)
	local info = self.lookup[spell]
	if (info) then
		if (info.exclusive) then
			self:ModifyTemplate(info.type, true)
			self:AddLimitedSpell(target, info.type)			-- Always add for things like Earth Shield, Fear Ward

		elseif (info.limited and (manual or listClick) and z:CanLearn()) then
			self:AddLimitedSpell(target, info.type)
		end
	end
end

-- AddLimitedSpell
function zg:AddLimitedSpell(name, key)
	if (not self.tickHandlers[key]) then
		error(format("No handler for %q", tostring(key)), 2)
	end

	local buff = self.buffs[key]
	if (not buff) then
		error(format("No self.buff for %q", key))
	end

	if (not template.limited) then
		template.limited = new()
	end

	if (buff.exclusive or not template.limited[key]) then
		del(template.limited[key])
		template.limited[key] = new()
	end

	if (not template.limited[key][name]) then
		template.limited[key][name] = true
		z:CheckForChange(self)
	end

	if (buff.exclusive) then
		self:AddSpellTracker(key, name)
	end
end

-- RemoveLimitedSpell
function zg:RemoveLimitedSpell(name, key)
	local part = template.limited and template.limited[key]
	if (part) then
		if (part[name]) then
			part[name] = nil
			z:CheckForChange(self)
		end
		
		if (not next(part) and self.buffs[key].exclusive) then
			self:StopSpellTracker(key)
		end

		if (self.db.char.resetOnClear) then
			if (not next(part)) then
				template.limited[key] = del(template.limited[key])
			end
			if (not next(template.limited)) then
				template.limited = del(template.limited)
			end
		end
	end
end

-- UNIT_SPELLCAST_FAILED
function zg:SpellCastFailed(spell, name, manual)
	if (not manual) then
		if (self.lookup[spell]) then
			z:Blacklist(name)
		end
	end
end

-- SPELLS_CHANGED
function zg:SPELLS_CHANGED()
	if (not z.zoneFlag) then
		self.actions = nil
		z:UpdateCellSpells()
		--self:GetMainSpells()
		self:MakeSpellOptions()

		--for name,spellDef in pairs(self.buffs) do
		--	if (spellDef.o == 1) then
		--		mainSingle = spellDef.list[1]
		--		mainGroup = spellDef.list[2]
		--		break
		--	end
		--end

		z:CheckForChange(self)
	end
end

-- SortedBuffList
function zg:SortedBuffList()
	local list = new()
	for Type,info in pairs(self.buffs) do
		local o = info.o
		local insertPos = #list + 1
		for i,k in ipairs(list) do
			if (self.buffs[k].o > o) then
				insertPos = i
				break
			end
		end
		tinsert(list, insertPos, Type)
	end
	return list
end

-- TooltipUpdate
function zg:TooltipUpdate(cat)
	if (template) then
		cat:AddLine('text', " ")
		cat:AddLine(
			"text", L["Group Template: "].."|cFFFFFFFF"..(self.db.char.selectedTemplate or L["none"]),
			"text2", (template and template.modified and "|cFFFF4040"..L["(modified)"].."|r") or ""
		)

		local list = self:SortedBuffList()
		for i,k in ipairs(list) do
			local key = self.buffs[k]
			if (GetSpellCooldown(key.list[1])) then
				local endis
				if (template[key.type]) then
					endis = "|cFF80FF80"..L["Enabled"].."|r"
				else
					endis = "|cFFFF8080"..L["Disabled"].."|r"
				end

				checkIcon = BS:GetSpellIcon(key.list[1])

				cat:AddLine(
					"text", ColourSpellFromKey(key),
					"text2", endis,
					"func", ToggleKeyType,
					"arg1", key.type,
					"hasCheck", checkIcon and true,
					"checked", checkIcon and true,
					"checkIcon", checkIcon
				)
			end
		end
		del(list)
	end
end

-- OnResetDB
function zg:OnResetDB()
	if (self.db) then
		template = self.db.char.templates.current

		for k,v in pairs(template) do
			if (k ~= "modified" and k ~= "state" and k ~= "limited") then
				if (not self.buffs[k]) then
					template[k] = nil
				end
			end
		end

		self:TickInitForTemplate()
	end
end

-- OnModuleEnable
function zg:OnModuleEnable()
	if (self.db) then
		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end

		z:MakeOptionsReagentList()
		self:OnResetDB()
		
		self:SPELLS_CHANGED()
		self:MakeSpellOptions()

		self:RegisterBucketEvent("UNIT_AURA", 0.2)				-- We don't care who
		self:RegisterBucketEvent("SPELLS_CHANGED", 2)
		z:CheckForChange(self)
	end
end

-- OnDisable
function zg:OnDisable()
	z:CheckForChange(self)
end
